"""
apkg packaging file cache
"""

import json
from pathlib import Path

from apkg.log import getLogger
from apkg.util.common import hash_file, hash_path


log = getLogger(__name__)


def file_checksum(*paths):
    return hash_file(*paths).hexdigest()[:20]


def path_checksum(*paths):
    return hash_path(*paths).hexdigest()[:20]


def enabled_str(enabled):
    return 'ENABLED' if enabled else 'DISABLED'


class ProjectCache:
    def __init__(self, project):
        self.project = project
        self.loaded = False
        self.cache = {}
        self.checksum = None

    def save(self):
        json.dump(self.cache, self.project.path.cache.open('w'))

    def load(self):
        cache_path = self.project.path.cache
        if not cache_path.exists():
            log.verbose("cache not found: %s", cache_path)
            return
        log.verbose("loading cache: %s", cache_path)
        self.cache = json.load(cache_path.open('r'))

    def _ensure_load(self):
        """
        ensure cache is loaded on demand and only once

        you don't need to call this directly
        """
        if self.loaded:
            return
        self.load()
        self.loaded = True

    def update(self, key, paths):
        """
        update cache entry
        """
        log.verbose("cache update: %s -> %s", key, paths[0])
        assert key
        self._ensure_load()
        entries = list(map(path2entry, paths))
        self.cache[key] = entries
        self.save()

    def get(self, key):
        """
        get cache entry or None
        """
        log.verbose("cache query: %s", key)

        def validate(path, checksum):
            if not path.exists():
                log.info("removing missing file from cache: %s", path)
                self.delete(key)
                return False
            real_checksum = file_checksum(path)
            if real_checksum != checksum:
                log.info("removing invalid cache entry: %s", path)
                self.delete(key)
                return False
            return True

        def entry2path_valid(e):
            return entry2path(e, validate_fun=validate)

        assert key
        self._ensure_load()
        entries = self.cache.get(key)
        if not entries:
            return None
        paths = list(map(entry2path_valid, entries))
        if None in paths:
            # invalid entry
            return None
        return paths

    def delete(self, key):
        """
        delete cache entry
        """
        self.cache.pop(key, None)
        self.save()

    def enabled(self, *targets, cmd=None, use_cache=True):
        """
        helper to tell and log if caching should be enabled

        targets is a list of cache targets that must be enabled:

        * local: cache local files
        * remote: cache remote files
        * source: cache project source (requires VCS)

        optional use_cache utility argument to disable all cache
        """

        def enabled_result(value):
            en_str = enabled_str(value)
            if cmd:
                log.verbose("cache %s for %s", en_str, cmd)
            else:
                log.verbose("cache %s", en_str)
            return value

        if not use_cache:
            # all cache disabled
            return enabled_result(use_cache)

        r = True
        for target in targets:
            option = 'cache.%s' % target
            value = self.project.config_get(option)
            cache = True

            if value is not None:
                # set in project config
                if value and target == 'source' and not self.project.vcs:
                    # source cache requires VCS
                    msg = ("cache.{target} {en} in project config, "
                           "but VCS isn't available - cache.{target} {di}.\n"
                           "Please update your project config.").format(
                        target=target,
                        en=enabled_str(value),
                        di=enabled_str(False))
                    log.warning(msg)
                    cache = False
                else:
                    log.verbose("cache.%s %s in project config",
                                target, enabled_str(value))
                    cache = value

            else:
                # auto cache settings
                if target == 'source':
                    # source cache requires VCS
                    cache = bool(self.project.vcs)
                    log.verbose("cache.%s %s by default (VCS=%s)",
                                target, enabled_str(cache), self.project.vcs)
                else:
                    # other cache types are ENABLED by default
                    log.verbose("cache.%s %s by default",
                                target, enabled_str(True))

            # always go through all targets to get complete logs
            r = r and cache

        return enabled_result(r)


def path2entry(path):
    """
    convert a path to corresponding cache entry

    return (fn, checksum) or a list of that on multiple paths
    """
    return str(path), file_checksum(path)


def entry2path(entry, validate_fun=None):
    """
    convert cache entry to a corresponding path

    if validate_fun is specified, it's used confirm file has
    valid checksum and flush invalid cache entry if it doesn't
    """
    fn, checksum = entry
    p = Path(fn)
    if validate_fun:
        if not validate_fun(p, checksum):
            return None
    return p



# hatchling uses this version when using python3 -m build
# setuptools read this version through setup.py
# scripts/make-archive.sh updates this using dunamai for `apkg make-archive`
__version__ = '0.5.1'



import click

from apkg import adistro
from apkg.cli import cli
from apkg.pkgstyle import call_pkgstyle_fun, get_pkgstyle_for_distro
from apkg.commands.get_archive import parse_archive_args
from apkg.util import common
from apkg.log import getLogger
from apkg.project import Project


log = getLogger(__name__)


@cli.command(name="test-dep", aliases=['testdep'])
@click.argument('input_files', nargs=-1)
@click.option('-l', '--list', 'install', default=True, flag_value=False,
              help="list test deps only, don't install")
@click.option('-u', '--upstream', is_flag=True,
              help="use upstream archive")
@click.option('-a', '--archive', is_flag=True,
              help="use test deps from archive")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('-F', '--file-list', 'input_file_lists', multiple=True,
              help=("specify text file listing one input file per line"
                    ", use '-' to read from stdin"))
@click.option('--ask/--no-ask', 'interactive',
              default=False, show_default=True,
              help="enable/disable interactive mode")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_test_dep(*args, **kwargs):
    """
    install or list testing dependencies
    """
    deps = test_dep(*args, **kwargs)
    if not kwargs.get('install', True):
        common.print_results(deps)


def test_dep(
        upstream=False,
        archive=False,
        input_files=None,
        input_file_lists=None,
        install=True,
        distro=None,
        interactive=False,
        project=None):
    """
    parse and optionally install testing dependencies

    pass install=False to only get list of deps without install

    returns list of test deps
    """
    action = 'installing' if install else 'listing'
    log.bold('%s testing deps', action)

    proj = project or Project()
    distro = adistro.distro_arg(distro, proj)
    log.info("target distro: %s", distro)

    infiles = common.parse_input_files(input_files, input_file_lists)
    archive, _ = parse_archive_args(
        proj, archive, upstream, infiles)

    tests = proj.get_tests_for_distro(distro)
    deps = tests.deps

    if install:
        n_deps = len(deps)
        if n_deps > 0:
            log.info("installing %s test deps...", n_deps)
            # TODO get pkgstyle for distro
            pkgstyle = get_pkgstyle_for_distro(distro)
            call_pkgstyle_fun(
                pkgstyle, 'install_build_deps',
                deps,
                distro=distro,
                interactive=interactive)
        else:
            log.info("no build deps to install")

    return deps


APKG_CLI_COMMANDS = [cli_test_dep]



# -*- encoding: utf-8 -*-

import re

import bs4
from packaging import version
import requests

from apkg.log import getLogger
from apkg.util.run import run


log = getLogger(__name__)


RE_ARCHIVE_VERSION = r'[\w-]+-(\d[^-]+)\.tar\..*'


def version_from_script(script, script_name='script'):
    """
    get version from last stdout line of a script
    """
    log.verbose("getting upstream version from %s: %s", script_name, script)
    out = run(script, quiet=True)
    _, _, last_line = out.rpartition('\n')
    v = version.parse(last_line.strip())
    return v


def version_from_listing(html_listing_url):
    """
    get latest version from HTML listing
    """
    log.verbose("getting upstream version from HTML listing: %s",
                html_listing_url)
    found = False
    v_max = version.parse('0')
    r = requests.get(html_listing_url)
    soup = bs4.BeautifulSoup(r.content, 'html.parser')
    for a in soup.find_all('a'):
        m = re.match(RE_ARCHIVE_VERSION, a.string)
        if not m:
            continue
        v_str = m.group(1)
        try:
            v = version.parse(v_str)
        except version.InvalidVersion:
            log.verbose("Ignoring invalid upstream version: %s in %s",
                        v_str, a.string)
            continue
        v_max = max(v, v_max)
        found = True
    if found:
        return v_max
    return None


def version_from_pypi(name):
    """
    get latest version from PyPI by name
    """
    log.verbose("getting upstream version from PyPI: %s", name)
    url = 'https://pypi.org/pypi/%s/json' % name
    r = requests.get(url)
    if not r.ok:
        return None
    data = r.json()

    versions = data['releases'].keys()
    ver = sorted(versions, key=version.parse)[-1]

    return ver


def latest_apkg_version():
    return version_from_pypi('apkg')



#!/usr/bin/env python3
"""
get latest upstream version from PyPI

This script can be used by apkg to check for latest upstream version
using upstream.version_script config option.
"""
from apkg.util.upstreamversion import version_from_pypi

# apkg expects last stdout line to contain the upstream version string
print(version_from_pypi('apkg'))



from contextlib import contextmanager
import fnmatch
import hashlib
from pathlib import Path
import sys
import tempfile

from apkg import ex
from apkg.log import getLogger
import apkg.util.shutil35 as shutil


log = getLogger(__name__)


def copy_paths(paths, dst):
    """
    utility to copy a list of paths to dst
    """
    if not dst.exists():
        dst.mkdir(parents=True, exist_ok=True)
    dst_full = dst.resolve()
    new_paths = []
    for p in paths:
        if p.parent.resolve() == dst_full:
            new_paths.append(p)
        else:
            p_dst = dst / p.name
            log.verbose("copying file: %s -> %s", p, p_dst)
            shutil.copy(p, p_dst)
            new_paths.append(p_dst)
    return new_paths


def get_cached_paths(proj, cache_key, result_dir=None):
    """
    get cached files and move them to result_dir if specified
    """
    paths = proj.cache.get(cache_key)
    if not paths:
        return None
    if result_dir:
        paths = copy_paths(paths, Path(result_dir))
    return paths


def print_results(results):
    """
    print results received from apkg command
    """
    try:
        for r in results:
            print(str(r))
    except TypeError:
        print(str(results))


def parse_input_files(files, file_lists):
    """
    utility to parse apkg input files and input file lists
    into a single list of input files
    """
    if not files:
        files = []
    if not file_lists:
        file_lists = []

    all_files = [Path(f) for f in files]

    if len([fl for fl in file_lists if fl == '-']) > 1:
        fail = "requested to read stdin multiple times"
        raise ex.InvalidInput(fail=fail)

    for fl in file_lists:
        if fl == '-':
            f = sys.stdin
        else:
            f = open(fl, 'r', encoding='utf-8')
        all_files += [Path(ln.strip()) for ln in f.readlines()]
        f.close()

    return all_files


def ensure_input_files(infiles):
    if not infiles:
        raise ex.InvalidInput(
            fail="no input file(s) specified")
    for f in infiles:
        if not f or not f.exists():
            raise ex.InvalidInput(
                fail="input file not found: %s" % f)


@contextmanager
def text_tempfile(text, prefix='apkg_tmp_'):
    """
    write text to a new temporary file and return its path

    file is deleted after use
    """
    f = tempfile.NamedTemporaryFile(
        prefix=prefix, mode='w+t', delete=False)
    path = Path(f.name)
    f.write(text)
    f.close()
    try:
        yield path
    finally:
        path.unlink()


def hash_file(*paths, algo='sha256'):
    """
    return hashlib's hash computed over the contents of the specified file

    typical use case: `hash_file('/path').hexdigest()`
    """
    # code based on https://stackoverflow.com/a/44873382/587396
    h = getattr(hashlib, algo)()
    b = bytearray(128*1024)
    mv = memoryview(b)
    for path in paths:
        # NOTE(py35): explicit Path -> str conversion for python 3.5
        with open(str(path), 'rb', buffering=0) as f:
            while True:
                # NOTE: pylint's cell-var-from-loop cries made me do this >:(
                n = f.readinto(mv)
                if n == 0:
                    break
                h.update(mv[:n])
    return h


def hash_path(*paths, algo='sha256'):
    """
    return hashlib's hash computed over the supplied file paths (as strings)

    typical use case: `hash_path('/path').hexdigest()`
    """
    h = getattr(hashlib, algo)()
    for path in paths:
        h.update(str(path).encode('utf-8'))
    return h


def fnmatch_any(filename, patterns):
    """
    check if a filename matches any of supplied patterns
    """
    for p in patterns:
        if fnmatch.fnmatch(filename, p):
            return True
    return False


class SortReversor:
    """
    use this with multi-key sort() to reverse individual keys
    """
    def __init__(self, obj):
        self.obj = obj

    def __eq__(self, other):
        return other.obj == self.obj

    def __lt__(self, other):
        return other.obj < self.obj



import json
import toml

import click
import distro as distro_mod

from apkg import adistro
from apkg.pkgstyle import PKGSTYLES
from apkg import pkgtemplate
from apkg.log import getLogger, T
from apkg.project import Project


log = getLogger(__name__)


@click.group(name='info')
@click.help_option('-h', '--help', help='show command help')
def cli_info():
    """
    show various apkg information
    """


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def cache():
    """
    show apkg cache contents
    """
    proj = Project()
    cache_str = "{t.bold}{fn}{t.normal}".format(fn=proj.path.cache, t=T)
    if proj.path.cache.exists():
        log.info("apkg cache: %s", cache_str)
        cdata = json.load(proj.path.cache.open('rt'))
        print(json.dumps(cdata, indent=4))
    else:
        log.info("apkg cache doesn't exist: %s", cache_str)


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def config():
    """
    show apkg project configuration
    """
    proj = Project()
    config_str = "{t.bold}{fn}{t.normal}".format(fn=proj.path.config, t=T)
    if proj.path.config.exists():
        log.info("project config: %s\n", config_str)
    else:
        log.info("project config doesn't exist: %s", config_str)

    if proj.config:
        print(toml.dumps(proj.config))


# pylint: disable=redefined-outer-name
@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def distro():
    """
    show current distro information
    """
    info = distro_mod.info()
    print(toml.dumps(info))


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def distro_aliases():
    """
    list available distro aliases
    """
    proj = Project()
    if not proj.distro_aliases:
        log.info("no distro aliases defined")
        return

    for name, al in proj.distro_aliases.items():
        msg = "{t.bold}{name}{t.normal}: {rules}"
        print(msg.format(name=name, rules=al, t=T))


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def pkgstyles():
    """
    list available packaging styles
    """
    for name, mod in PKGSTYLES.items():
        print("{t.bold}{name}{t.normal}".format(name=name, t=T))
        msg = "    module:   {t.magenta}{module}{t.normal}"
        print(msg.format(module=mod.__name__, t=T))
        msg = "    file:     {t.magenta}{fn}{t.normal}"
        print(msg.format(fn=mod.__file__, t=T))

        msg = "    distros:  "
        ds = ['{t.bold}%s{t.normal}' % d for d in mod.SUPPORTED_DISTROS]
        msg += ' | '.join(ds)
        print(msg.format(t=T))


@cli_info.command()
@click.option('-d', '--distro',
              help="set target distro  [default: current]")
@click.option('-c', '--custom', is_flag=True,
              help="only show custom variables per source")
@click.help_option('-h', '--help', help='show command help')
def template_variables(distro=None, custom=False):
    """
    show variables available in packaging template
    """
    proj = Project()
    distro = adistro.distro_arg(distro, proj)
    log.info("target distro: %s", distro)
    template = proj.get_template_for_distro(distro)
    if not custom:
        tvars = {'distro': distro}
        tvars = template.template_vars(tvars)
        print(toml.dumps(tvars))
        return

    # custom variables
    tvars = pkgtemplate.DUMMY_VARS
    tvars['distro'] = distro
    for vsrc in proj.variables_sources:
        print("# variables from %s: %s" % (vsrc.src_attr, vsrc.src_val))
        custom_tvars = vsrc.get_variables(tvars)
        print(toml.dumps(custom_tvars))
        tvars.update(custom_tvars)


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def upstream_version():
    """
    show detected project upstream version
    """
    proj = Project()
    msg = "upstream version: {t.bold}{v}{t.normal}"
    print(msg.format(v=proj.upstream_version, t=T))


APKG_CLI_COMMANDS = [cli_info]
